// -*- mode: c++; indent-tabs-mode: nil -*-
//! @file WSDL.qm WSDL: Web Services Description Language: http://www.w3.org/TR/wsdl

/*  WSDL.qm Copyright (C) 2012 - 2017 Qore Technologies, s.r.o.

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

// make sure we have the required qore version

// requires XML functionality

// need mime definitions

// do not use $ for vars



/*
    WSDL classes
    provides some minimal WSDL and XSD support for SOAP messaging used by the SoapClient class and the SoapHandler

    not complete, needs namespace verification, improved XSD support, element groups, etc
*/

/** @mainpage WSDL Module

    @tableofcontents

    @section wsdlintro Introduction to the WSDL Module

    The %WSDL module provides functionality for <a href="http://en.wikipedia.org/wiki/SOAP">SOAP</a> 1.1 and 1.2.

    The main purpose of this module is to provide the infrastructure for developing SOAP client and server services.

    The main components of this module are as follows:
    - @ref WSDL::WebService "WebService": represents a WSDL file; an easy way to get an object of this type from a URL is by calling WSDL::WSDLLib::getWSDL()
    - @ref WSDL::WSDLLib "WSDLLib": a class of static functions providing helper functions for the module
    - @ref WSDL::WSOperation "WSOperation": a class representing a single operation from a web service

    See the following for modules using the %WSDL module:
    - <a href="../../SoapClient/html/index.html">SoapClient user module</a>
    - <a href="../../SoapHandler/html/index.html">SoapHandler user module</a>

    @section wsdlrelnotes WSDL Module Release Notes

    @subsection wsdl_0_3_5_1 WSDL v0.3.5.1
    - supress emitting a SOAPAction header in requests if the binding gives an empty string (<a href="https://github.com/qorelanguage/qore/issues/1226">issue 1226</a>)
    - updated @ref WSDL::WSOperation::serializeRequest() to allow the SOAPAction header to be overridden in each request (<a href="https://github.com/qorelanguage/qore/issues/1226">issue 1226</a>)
    - respect XML generation flags in request generation
    - fixed parsing empty base64Binary and hexBinary elements (<a href="https://github.com/qorelanguage/qore/issues/1227">issue 1227</a>)

    @subsection wsdl_0_3_5 WSDL v0.3.5
    - fixed many message serialization and deserialization issues
    - added @ref WSDL::WebService::getOperation()
    - allow for environment variable substitution in WSDLLib::getWSDL() when retrieving files
    - fixed charset=... header value

    @subsection wsdl_0_3_4 WSDL v0.3.4
    - updated to a user module

    @subsection wsdl_0_3_3 WSDL v0.3.3
    - added initial support for the anyAttribute element of complexType
    - added initial support for SOAP header processing
    - added initial support for multiple portType and bindings in a WSDL

    @subsection wsdl_0_3_2 WSDL v0.3.2
    - fixed bugs de/serializing negative values for "int" and "short"

    @subsection wsdl_0_3_1 WSDL v0.3.1
    - improved XSD imports and namespace handling

    @subsection wsdl_0_3_0 WSDL v0.3.0
    - implemented WSDLLib class of helper functions
    - implemented support for xsd import statements in WSDLs

    @subsection wsdl_0_2_0 WSDL v0.2.0
    - use parseXMLAsData() instead of parseXML()
    - implemented initial simpleType support
    - fixed xsd:date serialization and deserialization
*/

//! main WSDL namespace
namespace WSDL {
    //! this WSDL implementation version
    const version     = "0.3.5.1";

    //! SOAP 1.1 envelope URI
    const SOAP_11_ENV  = "http://schemas.xmlsoap.org/soap/envelope/";
    //! SOAP 1.2 envelope URI
    const SOAP_12_ENV  = "http://www.w3.org/2003/05/soap-envelope";

    //! SOAP 1.1 namespace URI
    const SOAP_11_NS   = "http://schemas.xmlsoap.org/wsdl/soap/";

    //! SOAP 1.2 namespace URI
    const SOAP_12_NS   = "http://schemas.xmlsoap.org/wsdl/soap12/";

    //! XSD namespace URI
    const XSD_NS       = "http://www.w3.org/2001/XMLSchema";
    //! XSI namespace URI
    const XSI_NS       = "http://www.w3.org/2001/XMLSchema-instance";

    //! HTTP namespace URI
    const HTTP_NS      = "http://schemas.xmlsoap.org/wsdl/http/";
    //! MIME namespace URI
    const MIME_NS      = "http://schemas.xmlsoap.org/wsdl/mime/";

    //! soap 1.1 envelope namespaces
    const ENVELOPE_11_NS = (
        "soapenv:Envelope": (
            "^attributes^": (
                "xmlns:soapenv": SOAP_11_ENV,
                "xmlns:xsd": XSD_NS,
                "xmlns:xsi": XSI_NS,
            ),
        ),
        );

    //! soap 1.2 envelope namespaces
    const ENVELOPE_12_NS = (
        "soapenv:Envelope": (
            "^attributes^": (
                "xmlns:soapenv": SOAP_12_ENV,
              "xmlns:xsd": XSD_NS,
              "xmlns:xsi": XSI_NS,
            ),
        ),
        );

    //! soap encoding URI
    const SOAP_ENCODING = "http://schemas.xmlsoap.org/soap/encoding/";

    //! mapping from Qore types to xsd types for xsd type "anyType"
    const any_type_map = (
        Type::String      : "string",
        Type::Int         : "long",
        Type::Boolean     : "boolean",
        Type::Date        : "dateTime",
        Type::Float       : "decimal",
        Type::NothingType : "string",
        Type::NullType    : "string",
        Type::Binary      : "base64Binary",
        );

    // error codes
    const SOAP_SERIALIZATION_ERROR = "SOAP-SERIALIZATION-ERROR";
    const SOAP_DESERIALIZATION_ERROR = "SOAP-DESERIALIZATION-ERROR";
    const WSDL_ERROR = "WSDL-ERROR";

    //! SOAP HTTP transport URI
    const SOAP_TRANSPORT_HTTP = "http://schemas.xmlsoap.org/soap/http";

    //! known/supported transports
    const SOAP_TRANSPORT = (
        SOAP_TRANSPORT_HTTP: True,
        );

    //! range of "short" values (16 bits)
    const RANGE_SHORT = (-32768, 32767);

    //! range of "int" values (32 bits)
    const RANGE_INT = (-2147483648, 2147483647);
};

//! contains helper methods for retrieving WSDLs from a URL
class WSDL::WSDLLib {

public:
    //! Mime types recognized as SOAP messages
    const SoapMimeTypes = (MimeTypeSoapXml, MimeTypeXml, MimeTypeXmlApp);

    //! retrieves a local file and returns the file's contents as a string
/** called by WSDLLib::getFileFromURL() in case the scheme is "file"
    */
      static data getFile(string fn, bool as_string = True);

    //! retrieves a file from a URL with HTTP and returns the file's contents as a string
/** called by WSDLLib::getFileFromURL() in case the scheme is "http"
    */
      static string getHTTP(string url, __7_ string path, __7_ HTTPClient hc, __7_ hash headers, bool as_string = True);

    //! retrieves a file from a URL with the FTP protocol and returns the file's contents as binary or string data
/** called by WSDLLib::getFileFromURL() in case the scheme is "ftp"
    */
      static data getFTP(string url, string path, bool as_string = True);

    //! retrieves a file from a URL
/** in case the URL is actually a file path, environment variable substitution
        is performed on the path string
     */
    static data getFileFromURL(string url, string def_protocol = "file", __7_ HTTPClient http_client, __7_ hash http_headers, bool as_string = True, __7_ string def_path, __7_ reference new_def_path);


    //! retrieves a file from an already-parsed URL
/** in case the URL is actually a file path, environment variable substitution
        is performed on the path string
     */
    static data getFileFromURL(string url, hash u, string def_protocol = "file", __7_ HTTPClient http_client, __7_ hash http_headers, bool as_string = False, __7_ string def_path, __7_ reference new_def_path);


    //! returns the argument
      static WebService getWSDL(WebService wsdl);

    //! returns a WSDL string form a file name, optional HTTPClient object and optional header hash
/** in case the \a wsdl argument is actually a file path, environment variable substitution
        is performed on the path string
     */
      static string getWSDL(string wsdl, __7_ HTTPClient http_client, __7_ hash http_headers, __7_ reference new_def_path);

    //! takes a hash representation of a SOAP message and parses it to a Qore data structure; handles multipart messages, checks the content-type, and handles hrefs in the message
      static hash parseSOAPMessage(hash msg);


private:
      static checkSOAPContentType(string ct);
public:



private:
      static processHref(reference xmldata, string hr, hash parts);
public:



private:
      static substHref(reference xmldata, hash parts);
public:

};

//! base class with helper methods for XSD data processing
class WSDL::XsdBase {

public:
    /*
    static private hash doType(string t) {
        #printf("DEBUG: XsdBase::doType(%y)\n", t);
        (__7_ string ns, __7_ string type) = t =~ x/(\w+):(\w+)/;
        return !exists type ? ("val": t) : ("ns": ns, "val": type);
    }
    */

      static removeNS(reference v);

      static removeNS2(reference v);
};

//! base class for XSD data classes
class WSDL::XsdData : public WSDL::XsdBase {

public:
    any getValue(__7_ hash mrh, any val);

};

//! base class for XSD classes with a "name" attribute
class WSDL::XsdNamedData : public WSDL::XsdData {

public:
    public :
        // name of object
        string name;
        //! input namespace prefix (if any given)
        __7_ string ns;
        //! descriptive name flag
        descriptive_name;

public:

    constructor(string n_name, string n_ns);


    constructor(reference e, __7_ string desc_name);


    string getName();


    __7_ string getInputNamespacePrefix();


    bool hasRealName();

};

class WSDL::XsdAbstractType : public WSDL::XsdNamedData {

public:
    public :
        //! reference to namespaces
        Namespaces nsc;

        //! namespace output prefix
        string ons;

public:

    constructor(reference e, Namespaces n_nsc, __7_ string desc_name) ;


    constructor(string name, string ns, Namespaces n_nsc) ;


    private resolveNamespace();


    checkExtends(XsdAbstractType t, string ename);


    string getNameWithNS();


    bool isNillable();


    bool isRequired();


    bool requiresValue();


    string getOutputNamespacePrefix();


    abstract any serialize(any val, __7_ softbool omit_type);
    abstract any deserialize(string en, __7_ TypeMap tmap, __7_ hash mrh, any val);
};

//! class for XSD base types
class WSDL::XsdBaseType : public WSDL::XsdAbstractType {

public:
    public :

public:

    constructor(string t, Namespaces nsc, string ns = "xsd") ;


    any serialize(any val, __7_ softbool omit_type);


    any deserialize(string en, __7_ TypeMap tmap, __7_ hash mrh, any val);

};

//! class for XSD array types; currently only supports "binary"; used, for example with HTTP MultiPart messages
class WSDL::XsdArrayType : public WSDL::XsdAbstractType {

public:
    constructor(string t, Namespaces nsc) ;


    any serialize(any val, __7_ softbool omit_type);


    any deserialize(string en, __7_ TypeMap tmap, __7_ hash mrh, any val);

};

//! XSD typed data class
class WSDL::XsdTypedData : public WSDL::XsdNamedData {

public:
    public :
        any type;

public:

    constructor(reference e) ;

};

//! XSD attribute class
class WSDL::XsdAttribute : public WSDL::XsdTypedData {

public:
    public :
        string use = "optional";

        const AllowedUseValues = (
            "optional": True,
            "required": True,
            "prohibited": True,
            );

public:

    constructor(hash attr, Namespaces nsc, __7_ XsdAbstractType n_type, XsdLateResolverHelper unresolved) ;


    any getValue(string val);

};

//! XSD element class
class WSDL::XsdElement : public WSDL::XsdTypedData {

public:
    public :
        int minOccurs = 1;
        int maxOccurs = 1;
        bool nillable = False;
        __7_ string ref;
        bool usedocns;

public:

    constructor(hash e, Namespaces nsc, __7_ XsdAbstractType n_type, XsdLateResolverHelper unresolved, bool n_usedocns) ;


    assimilate(WSDL::XsdElement other);


    bool isRequired();


    bool isNillable();


    /*
    bool isRequired() {
        #printf("DBG %y: required nillable: %y minOccurs: %y type: %y\n", name, nillable, minOccurs, ((type instanceof XsdAbstractType && type.isRequired()) | True));
        return !nillable && minOccurs && ((type instanceof XsdAbstractType && type.isRequired()) | True);
    }
    */

    any serialize(any h, __7_ softbool omit_type, string key, string typename);


    private any serializeAsIntern(XsdAbstractType type, any h, __7_ softbool omit_type, string key, string typename);


    any deserialize(__7_ TypeMap tmap, __7_ hash mrh, any val, bool present);

};

//! XSD simple type class
class WSDL::XsdSimpleType : public WSDL::XsdAbstractType {

public:
    public :
        hash enum;
        any type;
        bool usedocns;

public:

    constructor(hash st, Namespaces nsc, XsdLateResolverHelper unresolved, bool n_usedocns, __7_ string desc_name) ;


    any serialize(any val, __7_ softbool omit_type);


    any deserialize(string en, __7_ TypeMap tmap, __7_ hash mrh, any val);

};

//! XSD complex type class
class WSDL::XsdComplexType : public WSDL::XsdAbstractType {

public:
    public :
        any array;
        any restriction;
        __7_ string extension;

        bool usedocns;

        __7_ hash elementmap;
        bool anyAttribute = False;

        bool nillable = False;
        bool required = False;
        bool requires_value = False;

        // attributes
        hash attrs;

        // any annotated documentation
        __7_ string documentation;

        // simpleContent type
        any simpleType;

        // finalization flag
        bool finalized = False;

        // multiple choice blocks
        list choices = ();

public:

    private :
        //! type of complexType object
        string cx_type;

        const XET_ALL      = "ALL";
        const XET_CHOICE   = "CHOICE";
        const XET_SEQUENCE = "SEQUENCE";
        const XET_SIMPLE   = "SIMPLE";
        const XET_NONE     = "NONE";

public:

    // ct can be NOTHING in case of an empty complex type
    constructor(__7_ hash ct, Namespaces nsc, XsdLateResolverHelper unresolved, bool n_usedocns, __7_ string desc_name) ;


    finalize(TypeMap tmap, Namespaces nsc);


    //! throws an exception if the types are not compatible
    checkExtends(XsdAbstractType t, string ename);


    private parseData(hash d, XsdLateResolverHelper unresolved);


    private parseAttributes(reference d, XsdLateResolverHelper unresolved);


    bool isNillable();


    bool isRequired();


    bool requiresValue();


    private __7_ hash parseElements(softlist el, XsdLateResolverHelper unresolved, bool for_object = True);


    private __7_ hash serializeElement(string key, XsdElement element, any h, __7_ softbool omit_type);


    __7_ hash serialize(any h, __7_ softbool omit_type);


    private hash serializeChoice(hash emap, hash h, __7_ softbool omit_type);


    __7_ hash deserialize(string en, __7_ TypeMap tmap, __7_ hash mrh, any oval);


    private __7_ hash parseChoice(hash val, hash emap, string en, __7_ TypeMap tmap, __7_ hash mrh, __7_ bool required);

};

//! web service operation class
class WSDL::WSOperation : public WSDL::XsdNamedData {

public:
    public :
        //! request message
        WSMessage input;
        //! response message
        WSMessage output;
        //! request message name
        __7_ string input_name;
        //! response message name
        __7_ string output_name;
        //! internal type map
        TypeMap tmap;

        //! namespace container
        Namespaces nsc;

        //! the SOAPAction header value to send with this operation
        __7_ string soapAction;

        //! the top-level request name
        __7_ string request_name;

        //! request multipart hash
/** this has has the following keys:
            - \c "multipart": the multipart type name (ex: \c "related")
            - \c "body": a hash of the XML attributes of the body declaration in the part
            - \c "parts": a hash keyed by part name with values of types for the part

	    @warning this member will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
         */
        __7_ hash inp;

        //! response multipart hash
/** this has has the following keys:
            - \c "multipart": the multipart type name (ex: \c "related")
            - \c "body": a hash of the XML attributes of the body declaration in the part
            - \c "parts": a hash keyed by part name with values of types for the part

	    @warning this member will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
         */
        __7_ hash outp;

        //! if the operation requires document style encoding
/**
	    @warning this member will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
         */
        bool docstyle = False;

        //! a hash with info about soap header requirements for request messages
/** the hash has the following keys:
            - \c "part": the part name to use for the header
            - \c "msg": a WSMessage object to use for serialization
            - \c "encoded": currently ignored

	    @warning this member will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
         */
        hash iheader;

        //! currently unused; reserved to be used to serialize header information for response messages
/**
	    @warning this member will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
	 */
        hash oheader;

public:

    //! creates the WSOperation object from the arguments supplied
/** @param p the raw hash from deserialized XML data for the operation
        @param n_tmap a reference to an object of a private class that stores XSD type information from types defined in the WSDL
        @param n_nsc the namespace container
        @param messages a hash of messages defined in the WSDL; keys are message names; values are WSMessage objects
     */
    constructor(hash p, TypeMap n_tmap, Namespaces n_nsc, __7_ hash messages) ;


    //! this method is called if this operation requires document style
    setDocStyle(reference idocmap);


    //! sets the top-level request name
    setTopLevelRequestElement(string name);


    //! returns the top-level request name
    string getTopLevelRequestName();


    //! serializes a request to an XML string for the operation
/** @param h the request to serialize
        @param header optional soap header info to serialize if required (ex: authorization info)
        @param enc the optional encoding to use; if this argument is not present, then the default encoding will be used
        @param nsh an optional namespace hash for the output message
        @param xml_opts optional XML generation options
        @param req_soapaction if present will override any SOAPAction value for the request

        @return a hash with keys:
        - \c body: XML string in the SOAP request format
        - \c hdr: hash of HTTP headers
        */
    hash serializeRequest(any h, __7_ hash header, __7_ string enc, __7_ hash nsh, __7_ int xml_opts, __7_ string req_soapaction);


    //! serializes a SOAP response to an XML string for the operation
/** @param h the response to serialize
        @param header SOAP header hash
        @param enc the optional encoding to use; if this argument is not present, then the default encoding will be used
        @param nsh namespace hash
        @param soap12 set to True if the response should use SOAP 1.2 encoding
        @param xml_opts optional XML generation options

        @return a hash with keys:
        - \c body: XML string in the SOAP request format
        - \c hdr: hash of HTTP headers
    */
    hash serializeResponse(any h, __7_ hash header, __7_ string enc, __7_ hash nsh, __7_ bool soap12, __7_ int xml_opts);


    //! processes multi-part references and returns a hash of multi-part reference info and the updated message body hash
    private list processMultiRef(hash body);


    //! parses a hash representing a parsed XML request (parsed with parseXMLAsData()) for the operation and returns the corresponding Qore data structure
/** @param o the parsed XML request (parsed with parseXMLAsData()) for the operation
        @return the Qore data structure corresponding to the request data
    */
    any deserializeRequest(hash o);


    //! parses a hash representing a parsed XML response (parsed with parse_xml()) for the operation and returns the corresponding Qore data structure
/** @param o the parsed XML response (parsed with parse_xml()) for the operation
        @return the Qore data structure corresponding to the response data
    */
    any deserializeResponse(hash o);


    //! processes a hash of raw deserialized XML data and removes any namespace prefix from keys (saved to a \c "ns" key in the hash value assigned to the key) and returns the processed version of the hash
      static private hash processNSValue(hash h);

    //! returns True if the operation is a SOAP 1.2 operation
/** @return True if the operation is a SOAP 1.2 operation
    */
    bool isSoap12();


    //! returns the target namespace for the operation
/** @return the target namespace for the operation
    */
    string getTargetNS();


    //! designates the operation as requiring multi-part output (responses) and sets the multipart output type
/**
        @warning this method will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
     */
    setOutputMultipart(string v);


    //! parses a message part from the raw deserialized hash representing the source XML for the part
/**
        @warning this method will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
     */
    private parsePart(reference msg, hash part);


    //! adds an output (response) part; the operation must be already flagged as requiring multi-part output
/** @param part the raw deserialized hash from the XML describing the part

        @warning this method will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
     */
    addOutputPart(hash part);


    //! designates the operation as requiring multi-part input (requests) and sets the multipart input type
/**
        @warning this method will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
     */
    setInputMultipart(string v);


    //! adds a request message part; the operation must be already flagged as requiring multi-part input (requests)
/** @param part the raw deserialized hash from the XML describing the part

        @warning this method will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
     */
    addInputPart(hash part);


    //! sets SOAP header information for the SOAP envelope for requests with this operation
/**
        @warning this method will disappear in the next major version of the WSDL module when support for multiple WSDL bindings and proper SOAP header processing is implemented
     */
    setInputHeader(string part, WSMessage msg, bool encoded);

};

//! web service message class
class WSDL::WSMessage : public WSDL::XsdNamedData {

public:
    public :
        hash args;
        // part map; maps element names to part names
        hash pmap;
        bool encoded = False;

        // keep a reference to Namespaces
        Namespaces nsc;

        // keep a reference to the type map
        TypeMap tmap;

public:

    constructor(hash m, ElementMap emap, TypeMap n_tmap, Namespaces n_nsc) ;


    hash serialize(any msginfo, any mpm, string n_name, any h);


    hash serializeDocument(__7_ string k, any msginfo, any mpm, any h);


    any deserialize(__7_ hash mrh, hash val);


    any deserializeDocument(__7_ hash mrh, any val);

};

// private helper class for lazy name resolution
class WSDL::XsdLateResolverHelper {

public:
private:
 list l = (); 
public:

    constructor();


    add(any v);


    list getList();


    clearResolved();


    bool isResolved(XsdSimpleType t);


    bool isResolved(XsdTypedData t);


    bool isResolved(XsdComplexType t);


    bool isResolved(any t);

};

//! class for WSDL bindings
class WSDL::Binding : public WSDL::XsdNamedData {

public:
    public :
        bool docStyle = False;

public:

    private :
        string port;

public:

    constructor(hash data, Namespaces nsc, hash opmap, reference idocmap, __7_ hash messages) ;


    string getPort();

};

// private namespace prefix redefinition class
class WSDL::NamespacePrefixHelper {

public:
    public :

public:

    private :
        Namespaces nsc;

        // overriden prefixes
        hash h;

        // overridden target namespace
        bool targ_ns;

        // overridden default namespace
        bool def_ns;

public:

    constructor(Namespaces n_nsc, __7_ hash nsattr);


    destructor();


    save(string k, string v);

};

//! namespace container class
class WSDL::Namespaces {

public:
    public :

public:

    private :
        //! maps output namespace prefixes to namespace URIs
        hash ns = (
            "xsd": XSD_NS,
            );

        //! maps namespace URIs to output namespace prefixes
        hash nsr = (
            XSD_NS: "xsd",
            );

        //! hash for valid XSD namespaces, maps input namespace prefixes -> True if it refers to an XSD schema
        hash xsd_schema = (
            "xsd": True,
            );

        //! hash mapping input namespace prefixes to namespaces URIs
        hash imap = (
            "xsd": XSD_NS,
            );

        //! hash mapping input namespace URIs to input namespace prefixes
        hash imapr = (
            XSD_NS: "xsd",
            );

        //! integer providing a sequence for output namespace prefixes
        int nsn;

        //! if True then has SOAP 1.1
        bool hassoap11 = False;

        //! if True then has SOAP 1.2
        bool hassoap12 = False;

        //! current target namespace
        __7_ string target_ns;

        //! target namespace stack;
        list nss = ();

        //! default namespace for unprefixed definitions
        __7_ string default_ns;

        //! default namespace stack;
        list dss = ();

public:

    //! creates the object with the WSDL definitions attribute hash
    constructor(hash nsh, __7_ Namespaces nsc);


    __7_ string getDefaultNs();


    addNamespaces(hash nsh, __7_ NamespacePrefixHelper nph);


    restorePrefixes(hash h);


    private addNamespaceIntern(string ns, string val, __7_ bool override);


    //! merges namespaces when parsing imported schemas
    merge(Namespaces nsc);


    //! returns the namespace URI for the given output namespace prefix
    string getOutputNamespaceUri(string nsp);


    //! returns a hash of namespace attributes for outgoing SOAP messages
    __7_ hash getOutputNamespaceHash(__7_ hash nsh);


    //! returns the primary target namespace Uri
    __7_ string getTargetNamespaceUri();


    //! returns the namespace URI for the given prefix or the target namespace Uri
    __7_ string getNamespaceUri(__7_ string nsp);


    //! pushes the current target namespace URI on the stack when parsing schemas and sets the current target namespace URI to the current value
    pushTargetNamespace(string ns);


    //! restores any previous target namespace URI from the stack to the current target namespace URI
    popTargetNamespace();


    //! pushes the current default namespace URI on the stack when parsing schemas and sets the current default namespace URI to the current value
    pushDefaultNamespace(string ns);


    //! restores any previous default namespace URI from the stack to the current default namespace URI
    popDefaultNamespace();


    //! looks up and registers a namespace if necessary, returns the namespace's prefix
    string getOutputNamespacePrefix(string ns);


    //! looks up and registers a namespace if necessary, returns the namespace's prefix
    string getTargetNamespaceInputPrefix();


    //! registers a namespace internally
    private string registerNamespaceIntern(string n_ns);


    //! returns a hash of namespace prefixes to namespaces URIs actually used
    __7_ hash getReferencedNamespaceMap();


    //! returns True if using SOAP 1.1, False if not
    bool hasSoap11();


    //! returns True if using SOAP 1.2, False if not
    bool hasSoap12();


    //! returns True if if the input namespace prefix refers to the XSD namespace URI
    __7_ bool isSchema(string ns);


    //! returns the input namespace URI from the input namespace prefix
    string getInputNamespaceUri(string nsa);


    any doType(string t);


    //! returns the output namespace prefix for the given input namespace prefix
    string translateOutputNamespacePrefix(__7_ string nsa);

};

// private class
class WSDL::MapBase {

public:
private:

public:

    private :
        hash dh;

public:

    list keys();


    AbstractIterator iterator();

};

// private class
class WSDL::TypeMap : public WSDL::MapBase {

public:
    add(XsdAbstractType t);


    XsdAbstractType get(string name);


    __7_ XsdAbstractType tryGet(string name);

};

// private class
class WSDL::ElementMap : public WSDL::MapBase {

public:
    add(XsdElement e);


    XsdElement get(string ns, string name);


    __7_ XsdElement tryGet(string ns, string name);

};

//! main class representing a parsed WSDL file
/** This is the main class for handling SOAP communication and is based on a WSDL file
*/
class WSDL::WebService : public WSDL::XsdBase {

public:
    public :
        //! the WSDL string
        string wsdl;

        //! namespace container
        Namespaces nsc;

        hash base_type;
        list wsdl_services = ();
        hash idocmap;
        hash opmap;
        ElementMap emap();
        hash messages;
        TypeMap tmap();
        bool usedocns = False;
        hash portType;
        __7_ code try_import;
        //! default path for retrieving XSD references
        __7_ string def_path;

public:

    private :
        // service definitions; name -> service hash
        hash services;

        // service bindings; name -> hash
        hash binding;

public:

    //! creates the WebService object
/** @param str the XML string representing the WSDL
        @param opts an optional hash of options with the following possible keys:
        - \c "def_path": the default path to use when retrieving referenced XSDs
        - \c "try_import": a call reference or closure to be passed a string name for XSD imports without any scheme to retrieve the data, must take a string argument (the resource name) and return a string (the resource data)
        - \c "http_client": a HTTPClient object for retrieving data from import commands
        - \c "http_headers": a hash of optional HTTP header info to use when retrieving data from import commands
    */
    constructor(string str, __7_ hash opts);


    //! returns the given operation or throws an exception if it cannot be found; operations are searched in portTypes in the order they are declared
    WSOperation getOperation(string opname);


    //! returns the given operation or throws an exception if it cannot be found
    WSOperation getOperation(string port, string opname);


    //! returns a list of hashes giving supported operation names for each port in the WSDL
/** @return a list of hashes giving supported operation names for each port in the WSDL; each hash has the following keys:
        - \c "port": the name of the port defining the operation
        - \c "operation": the @ref WSDL::WSOperation object for the operation
     */
    list listOperations();


    //! return a @ref WSDL::Binding object describing the requested binding
/** @param name the name of the binding

        @return a @ref WSDL::Binding object

        @throw WSDL-BINDING-ERROR unknown binding
     */
    WSDL::Binding getBinding(string name);


    //! returns a list of services defined in the WSDL
/** @return a list of hashes of services defined in the WSDL; each hash has the following keys:
        - \c "name": the name of the service
        - \c "port": a hash of port information; the keys are port names and the values have the following keys:
          - \c "address": the location of the port
          - \c "binding": the binding of the port
     */
    list listServices();


    //! returns a hash describing the requested service
/** @param name the name of the service
        @return a hash describing the service with the following keys:
        - \c "name": the name of the service
        - \c "port": a hash of port information; where each hash is keyed by port name and the values have the following keys:
          - \c "address": the location of the port
          - \c "binding": the binding of the port

        @throw WSDL-SERVICE-ERROR the service is not known
     */
    hash getService(string name);


    hash getType(string name, any v);


    private XsdBaseType getBaseType(string t);


    private resolveType(XsdSimpleType t);


    private resolveType(XsdElement xe);


    private resolveType(XsdTypedData xd);


    private resolveType(XsdComplexType ct);


    private XsdAbstractType resolveType(hash v);


    // parse XSD schema types
    private parseTypes(__7_ hash data, any http_client, any http_headers);


    private parseMessages(__7_ softlist message);


    private parseService(__7_ softlist svcs);


    private parsePortType(__7_ softlist data);


    private parseBinding(__7_ softlist bindings);


    //! returns True if the WSDL describes a SOAP 1.2 service
/** @return True if the WSDL describes a SOAP 1.2 service
    */
    bool isSoap12();


    //! returns the XML string for the WSDL
/** @return the XML string for the WSDL
    */
    string getWSDL();


    //! returns the XML string for the WSDL, adjusting the URLs for binding locations depending on the caller
/** @param base_url the url to be prepended before the service name;
                        it must include all the handler prefixes (like \c /SOAP)
                        to produce valid URL
        @return the XML string for the WSDL
     */
    string getWSDL(string base_url);

};
